# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Millan Castro <m.castrovilarino@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Markers display and management."""
import os
from typing import Optional

from gi.repository import Gdk
from gi.repository import GES
from gi.repository import Gtk

from pitivi.configure import get_ui_dir
from pitivi.utils.loggable import Loggable
from pitivi.utils.timeline import Zoomable

TIMELINE_MARKER_SIZE = 10
CLIP_MARKER_HEIGHT = 12
CLIP_MARKER_WIDTH = 10


class Marker(Gtk.EventBox, Loggable):
    """Widget representing a marker."""

    def __init__(self, ges_marker, class_name, width, height):
        Gtk.EventBox.__init__(self)
        Loggable.__init__(self)

        self.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK |
                        Gdk.EventMask.LEAVE_NOTIFY_MASK)

        self.ges_marker = ges_marker
        self.ges_marker.ui = self
        self.position_ns = self.ges_marker.props.position
        self.width = width
        self.height = height

        self.get_style_context().add_class(class_name)
        self.ges_marker.connect("notify-meta", self._notify_meta_cb)

        self._selected = False

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.CONSTANT_SIZE

    def do_get_preferred_height(self):
        return self.height, self.height

    def do_get_preferred_width(self):
        return self.width, self.width

    def do_enter_notify_event(self, unused_event):
        self.set_state_flags(Gtk.StateFlags.PRELIGHT, clear=False)

    def do_leave_notify_event(self, unused_event):
        self.unset_state_flags(Gtk.StateFlags.PRELIGHT)

    def _notify_meta_cb(self, unused_ges_marker, item, value):
        self.set_tooltip_text(self.comment)

    @property
    def position(self):
        """Returns the position of the marker, in nanoseconds."""
        return self.ges_marker.props.position

    @property
    def comment(self) -> str:
        """Returns a comment from ges_marker."""
        return self.ges_marker.get_string("comment") or ""

    @comment.setter
    def comment(self, text: str):
        if text == self.comment:
            return

        self.ges_marker.set_string("comment", text)

    @property
    def selected(self):
        """Returns whether the marker is selected."""
        return self._selected

    @selected.setter
    def selected(self, selected):
        self._selected = selected
        if self._selected:
            self.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)
        else:
            self.unset_state_flags(Gtk.StateFlags.SELECTED)


class MarkersBox(Gtk.EventBox, Zoomable, Loggable):
    """Container for displaying and managing markers."""

    def __init__(self, app, hadj=None):
        Gtk.EventBox.__init__(self)
        Zoomable.__init__(self)
        Loggable.__init__(self)

        self.layout = Gtk.Layout()
        self.add(self.layout)
        self.get_style_context().add_class("MarkersBox")

        self.app = app

        if hadj:
            hadj.connect("value-changed", self._hadj_value_changed_cb)
        self.props.hexpand = True
        self.props.valign = Gtk.Align.START

        self._offset = 0
        self.props.height_request = TIMELINE_MARKER_SIZE

        self.__markers_container: Optional[GES.MarkerList] = None
        self.marker_moving: Optional[Marker] = None
        self.marker_new: Optional[Marker] = None

        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)

    def first_marker(self, before: Optional[int] = None, after: Optional[int] = None) -> Optional[int]:
        """Returns position of the closest marker found before or after the given timestamp.

        None is returned if no such marker is found.
        """
        assert (after is not None) != (before is not None)

        if not self.markers_container:
            return None

        if after is not None:
            start = after + 1
            end = self.app.project_manager.current_project.ges_timeline.props.duration
        else:
            start = 0
            end = before

        if start >= end:
            return None

        markers_positions = list(ges_marker.props.position
                                 for ges_marker in self.markers_container.get_markers()
                                 if start <= ges_marker.props.position < end)
        if not markers_positions:
            return None

        if after is not None:
            return min(markers_positions)
        else:
            return max(markers_positions)

    def add_at_timeline_time(self, position):
        """Adds a marker at the given timeline position."""
        if not self.markers_container:
            return

        self.markers_container.add(position)

    @property
    def markers_container(self):
        """Gets the GES.MarkerList."""
        return self.__markers_container

    @markers_container.setter
    def markers_container(self, ges_markers_container):
        if self.__markers_container:
            for marker in self.layout.get_children():
                self.layout.remove(marker)
            self.__markers_container.disconnect_by_func(self._marker_added_cb)
            self.__markers_container.disconnect_by_func(self._marker_removed_cb)
            self.__markers_container.disconnect_by_func(self._marker_moved_cb)

        self.__markers_container = ges_markers_container
        if self.__markers_container:
            self.__create_marker_widgets()
            self.__markers_container.connect("marker-added", self._marker_added_cb)
            self.__markers_container.connect("marker-removed", self._marker_removed_cb)
            self.__markers_container.connect("marker-moved", self._marker_moved_cb)

    def release(self):
        if self.__markers_container:
            self.__markers_container.disconnect_by_func(self._marker_added_cb)
            self.__markers_container.disconnect_by_func(self._marker_removed_cb)
            self.__markers_container.disconnect_by_func(self._marker_moved_cb)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        if self.offset == value:
            return

        self._offset = value
        self._update_position()

    def __create_marker_widgets(self):
        markers = self.__markers_container.get_markers()

        for ges_marker in markers:
            position = ges_marker.props.position
            self._add_marker(position, ges_marker)

    def _hadj_value_changed_cb(self, hadj):
        """Handles the adjustment value change."""
        self.offset = hadj.get_value()

    def zoom_changed(self):
        self._update_position()

    def _update_position(self):
        for marker in self.layout.get_children():
            position = self.ns_to_pixel(marker.position) - self.offset - marker.width / 2
            self.layout.move(marker, position, 0)

    def do_button_press_event(self, event):
        if not self.markers_container:
            return False

        if event.button == Gdk.BUTTON_PRIMARY:
            event_widget = Gtk.get_event_widget(event)
            if isinstance(event_widget, Marker):
                if event.type == Gdk.EventType.BUTTON_PRESS:
                    self.marker_moving = event_widget
                    self.marker_moving.selected = True
                    self.app.action_log.begin("Move marker", toplevel=True)
                    return True

                if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                    self.marker_moving = None
                    self.app.action_log.rollback()
                    marker_popover = MarkerPopover(self.app, event_widget, self.__markers_container)
                    marker_popover.popup()
                    return True

            else:
                position = self.pixel_to_ns(event.x + self.offset)
                with self.app.action_log.started("Added marker", toplevel=True):
                    self.__markers_container.add(position)
                    self.marker_new.selected = True
                return True

        return False

    def do_button_release_event(self, event):
        if not self.markers_container:
            return False

        if event.button == Gdk.BUTTON_PRIMARY:
            if self.marker_moving:
                self.marker_moving.selected = False
                self.marker_moving = None
                self.app.action_log.commit("Move marker")
                return True

            if self.marker_new:
                self.marker_new.selected = False
                self.marker_new = None
                return True

        return False

    def do_motion_notify_event(self, event):
        if not self.markers_container:
            return False

        event_widget = Gtk.get_event_widget(event)
        if event_widget is self.marker_moving:
            event_x, unused_y = event_widget.translate_coordinates(self, event.x, event.y)
            event_x = max(0, event_x)
            position_ns = self.pixel_to_ns(event_x + self.offset)
            self.__markers_container.move(self.marker_moving.ges_marker, position_ns)
            return True

        return False

    def _marker_added_cb(self, unused_markers, position, ges_marker):
        self._add_marker(position, ges_marker)

    def _create_marker(self, ges_marker):
        return Marker(ges_marker, "Marker", TIMELINE_MARKER_SIZE, TIMELINE_MARKER_SIZE)

    def _add_marker(self, position, ges_marker):
        marker = self._create_marker(ges_marker)
        x = self.ns_to_pixel(position) - self.offset - marker.width / 2
        self.layout.put(marker, x, 0)
        marker.show()
        self.marker_new = marker

    def _marker_removed_cb(self, unused_markers, ges_marker):
        self._remove_marker(ges_marker)

    def _remove_marker(self, ges_marker):
        if not ges_marker.ui:
            return

        self.layout.remove(ges_marker.ui)
        ges_marker.ui = None

    def _marker_moved_cb(
            self, unused_markers, unused_prev_position, position, ges_marker):
        self._move_marker(position, ges_marker)

    def _move_marker(self, position, ges_marker):
        x = self.ns_to_pixel(position) - self.offset - ges_marker.ui.width / 2
        self.layout.move(ges_marker.ui, x, 0)


@Gtk.Template(filename=os.path.join(get_ui_dir(), "markerpopover.ui"))
class MarkerPopover(Gtk.Popover):
    """A popover to edit a marker's metadata or to remove the marker."""

    __gtype_name__ = "MarkerPopover"

    comment_textview = Gtk.Template.Child()
    remove_button = Gtk.Template.Child()

    def __init__(self, app, marker: Marker, markers_container: GES.MarkerList):
        Gtk.Popover.__init__(self)

        self.app = app
        self.marker: Optional[Marker] = marker
        self.markers_container: GES.MarkerList = markers_container

        self.comment_textview.get_buffer().set_text(self.marker.comment)

        self.set_position(Gtk.PositionType.TOP)
        self.set_relative_to(self.marker)
        self.show_all()

    @Gtk.Template.Callback()
    def remove_button_clicked_cb(self, event):
        with self.app.action_log.started("Removed marker", toplevel=True):
            self.markers_container.remove(self.marker.ges_marker)

        self.marker = None

        self.hide()

    def do_closed(self):
        if not self.marker:
            # The user clicked the Remove button so no need to update the text.
            return

        buffer = self.comment_textview.get_buffer()
        if buffer.props.text != self.marker.comment:
            with self.app.action_log.started("marker comment", toplevel=True):
                self.marker.comment = buffer.props.text
        self.marker.selected = False


class ClipMarkersBox(MarkersBox):
    """Bar at the top of a clip for displaying and editing clip markers."""

    def __init__(self, app, ges_elem: GES.Source, hadj: Optional[Gtk.Adjustment] = None):
        MarkersBox.__init__(self, app, hadj=hadj)

        self.ges_elem = ges_elem
        # We need a GESClip to convert timestamps.
        self.ges_clip = self.ges_elem.get_parent()

        self.props.height_request = CLIP_MARKER_HEIGHT

        self.get_style_context().add_class("ClipMarkersBox")

    def _create_marker(self, ges_marker):
        return Marker(ges_marker, "ClipMarker", CLIP_MARKER_WIDTH, CLIP_MARKER_HEIGHT)

    def __internal_to_timeline(self, timestamp):
        return self.ges_clip.get_timeline_time_from_internal_time(self.ges_elem, timestamp)

    def __timeline_to_internal(self, timestamp):
        return self.ges_clip.get_internal_time_from_timeline_time(self.ges_elem, timestamp)

    def first_marker(self, before: Optional[int] = None, after: Optional[int] = None) -> Optional[int]:
        assert (after is not None) != (before is not None)

        if not self.markers_container:
            return None

        # Limit search to visible markers
        clip_start = self.ges_clip.props.start
        clip_end = self.ges_clip.props.start + self.ges_clip.props.duration

        if after is not None:
            start = max(after + 1, clip_start)
            end = clip_end
        else:
            start = clip_start
            end = min(before, clip_end)

        if start >= end:
            return None

        markers_positions = [self.__internal_to_timeline(marker.props.position)
                             for marker in self.markers_container.get_markers()
                             if start <= self.__internal_to_timeline(marker.props.position) < end]
        if not markers_positions:
            return None

        if after is not None:
            return min(markers_positions)
        else:
            return max(markers_positions)

    def add_at_timeline_time(self, position):
        if not self.markers_container:
            return

        start = self.ges_clip.props.start
        inpoint = self.ges_clip.props.in_point
        duration = self.ges_clip.props.duration

        # Prevent timestamp conversion failing due to negative result.
        if position < start:
            return

        internal_end = self.__timeline_to_internal(start + duration)
        timestamp = self.__timeline_to_internal(position)

        # Check if marker would land in the 'visible' part of the clip.
        if not inpoint <= timestamp <= internal_end:
            return

        self.markers_container.add(timestamp)

    def do_button_press_event(self, event):
        """Ignore press events unless the clip is the only one selected."""
        if not self.ges_clip.selected:
            return False

        selection = self.app.gui.editor.timeline_ui.timeline.selection
        if selection.get_single_clip() is not self.ges_clip:
            return False

        return MarkersBox.do_button_press_event(self, event)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import os
from gettext import gettext as _
from time import time
from urllib.parse import unquote

from gi.repository import Gdk
from gi.repository import GES
from gi.repository import Gio
from gi.repository import Gtk

from pitivi.clipproperties import ClipProperties
from pitivi.configure import APPNAME
from pitivi.configure import get_ui_dir
from pitivi.dialogs.missingasset import MissingAssetDialog
from pitivi.dialogs.projectsettings import ProjectSettingsDialog
from pitivi.editorstate import EditorState
from pitivi.effects import EffectListWidget
from pitivi.interactiveintro import InteractiveIntro
from pitivi.mediafilespreviewer import PreviewWidget
from pitivi.medialibrary import MediaLibraryWidget
from pitivi.perspective import Perspective
from pitivi.settings import GlobalSettings
from pitivi.tabsmanager import BaseTabs
from pitivi.timeline.previewers import ThumbnailCache
from pitivi.timeline.timeline import TimelineContainer
from pitivi.transitions import TransitionsListWidget
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import path_from_uri
from pitivi.utils.ui import beautify_time_delta
from pitivi.utils.ui import EDITOR_PERSPECTIVE_CSS
from pitivi.utils.ui import info_name
from pitivi.viewer.viewer import ViewerContainer


GlobalSettings.add_config_section("main-window")
GlobalSettings.add_config_option('mainWindowHPanePosition',
                                 section="main-window",
                                 key="hpane-position",
                                 type_=int)
GlobalSettings.add_config_option('mainWindowMainHPanePosition',
                                 section="main-window",
                                 key="main-hpane-position",
                                 type_=int)
GlobalSettings.add_config_option('mainWindowVPanePosition',
                                 section="main-window",
                                 key="vpane-position",
                                 type_=int)
GlobalSettings.add_config_option('lastProjectFolder',
                                 section="main-window",
                                 key="last-folder",
                                 environment="PITIVI_PROJECT_FOLDER",
                                 default=os.path.expanduser("~"))


class EditorPerspective(Perspective, Loggable):
    """Pitivi's Editor perspective.

    Attributes:
        app (Pitivi): The app.
    """

    def __init__(self, app):
        Perspective.__init__(self)
        Loggable.__init__(self)

        self.app = app
        self.settings = app.settings

        self.builder = Gtk.Builder()
        self.editor_state = EditorState(app.project_manager)

        pm = self.app.project_manager
        pm.connect("new-project-loaded",
                   self._project_manager_new_project_loaded_cb)
        pm.connect("save-project-failed",
                   self._project_manager_save_project_failed_cb)
        pm.connect("project-saved", self._project_manager_project_saved_cb)
        pm.connect("closing-project", self._project_manager_closing_project_cb)
        pm.connect("reverting-to-saved",
                   self._project_manager_reverting_to_saved_cb)
        pm.connect("project-closed", self._project_manager_project_closed_cb)
        pm.connect("missing-uri", self._project_manager_missing_uri_cb)

    def setup_ui(self):
        """Sets up the UI."""
        self.__setup_css()
        self._create_ui()
        self.app.gui.connect("focus-in-event", self.__focus_in_event_cb)
        self.app.gui.connect("destroy", self._destroyed_cb)

    def activate_compact_mode(self):
        """Shrinks widgets to suit better a small screen."""
        self.medialibrary.activate_compact_mode()
        self.viewer.activate_compact_mode()

    def refresh(self):
        """Refreshes the perspective."""
        self.timeline_ui.restore_state()
        self.focus_timeline()

    def __setup_css(self):
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(EDITOR_PERSPECTIVE_CSS.encode("UTF-8"))
        screen = Gdk.Screen.get_default()
        style_context = self.app.gui.get_style_context()
        style_context.add_provider_for_screen(screen, css_provider,
                                              Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def __focus_in_event_cb(self, unused_widget, unused_event):
        ges_timeline = self.timeline_ui.timeline.ges_timeline
        if not ges_timeline:
            # Nothing to work with, Pitivi is starting up.
            return

        if self.app.project_manager.current_project.pipeline.rendering():
            return

        # Commit the timeline so its nested timelines assets are refreshed.
        ges_timeline.commit()

        # We need to track the changed assets ourselves.
        changed_files_uris = ThumbnailCache.update_caches()
        if changed_files_uris:
            self.medialibrary.update_asset_thumbs(changed_files_uris)

            for ges_layer in ges_timeline.get_layers():
                for ges_clip in ges_layer.get_clips():
                    if ges_clip.get_asset().props.id in changed_files_uris:
                        if ges_clip.ui.audio_widget:
                            ges_clip.ui.audio_widget.update_previewer()
                        if ges_clip.ui.video_widget:
                            ges_clip.ui.video_widget.update_previewer()

    def _destroyed_cb(self, unused_main_window):
        """Cleanup before destroying this window."""
        pm = self.app.project_manager
        pm.disconnect_by_func(self._project_manager_new_project_loaded_cb)
        pm.disconnect_by_func(self._project_manager_save_project_failed_cb)
        pm.disconnect_by_func(self._project_manager_project_saved_cb)
        pm.disconnect_by_func(self._project_manager_closing_project_cb)
        pm.disconnect_by_func(self._project_manager_reverting_to_saved_cb)
        pm.disconnect_by_func(self._project_manager_project_closed_cb)
        pm.disconnect_by_func(self._project_manager_missing_uri_cb)
        self.toplevel_widget.remove(self.timeline_ui)
        self.timeline_ui.destroy()

    def _render_cb(self, unused_button):
        """Shows the RenderDialog for the current project."""
        from pitivi.render import RenderDialog

        project = self.app.project_manager.current_project
        dialog = RenderDialog(self.app, project)
        dialog.window.show()

    def _create_ui(self):
        """Creates the graphical interface.

        The rough hierarchy is:
        vpaned:
        - mainhpaned(secondhpaned(main_tabs, context_tabs), viewer)
        - timeline_ui

        The full hierarchy can be admired by starting the GTK+ Inspector
        with Ctrl+Shift+I.
        """
        # pylint: disable=attribute-defined-outside-init
        # Main "toolbar" (using client-side window decorations with HeaderBar)
        self.headerbar = self.__create_headerbar()

        # Set up our main containers, in the order documented above

        # Separates the tabs+viewer from the timeline
        self.toplevel_widget = Gtk.Paned(orientation=Gtk.Orientation.VERTICAL)
        # Separates the tabs from the viewer
        self.mainhpaned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL)
        # Separates the two sets of tabs
        self.secondhpaned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL)
        self.toplevel_widget.pack1(self.mainhpaned, resize=False, shrink=False)
        self.mainhpaned.pack1(self.secondhpaned, resize=True, shrink=False)
        self.toplevel_widget.show()
        self.secondhpaned.show()
        self.mainhpaned.show()

        # First set of tabs
        self.main_tabs = BaseTabs(self.app)
        self.medialibrary = MediaLibraryWidget(self.app)
        self.effectlist = EffectListWidget(self.app)
        self.main_tabs.append_page("Media Library",
                                   self.medialibrary, Gtk.Label(label=_("Media Library")))
        self.main_tabs.append_page("Effect Library",
                                   self.effectlist, Gtk.Label(label=_("Effect Library")))
        self.medialibrary.connect('play', self._media_library_play_cb)
        self.medialibrary.show()
        self.effectlist.show()

        # Second set of tabs
        self.context_tabs = BaseTabs(self.app)
        self.clipconfig = ClipProperties(self.app)
        self.trans_list = TransitionsListWidget(self.app)
        self.context_tabs.append_page("Clip",
                                      self.clipconfig, Gtk.Label(label=_("Clip")))
        self.context_tabs.append_page("Transition",
                                      self.trans_list, Gtk.Label(label=_("Transition")))
        # Show by default the Title tab, as the Clip and Transition tabs
        # are useful only when a clip or transition is selected, but
        # the Title tab allows adding titles.
        self.context_tabs.set_current_page(2)

        self.secondhpaned.pack1(self.main_tabs, resize=False, shrink=False)
        self.secondhpaned.pack2(self.context_tabs, resize=False, shrink=False)
        self.main_tabs.show()
        self.context_tabs.show()

        # Viewer
        self.viewer = ViewerContainer(self.app)
        self.mainhpaned.pack2(self.viewer, resize=True, shrink=False)

        # Now, the lower part: the timeline
        self.timeline_ui = TimelineContainer(self.app, self.editor_state)
        self.toplevel_widget.pack2(self.timeline_ui, resize=True, shrink=False)

        self.intro = InteractiveIntro(self.app)
        self.headerbar.pack_end(self.intro.intro_button)

        # Setup shortcuts for HeaderBar buttons and menu items.
        self._create_actions()

        # Identify widgets for AT-SPI, making our test suite easier to develop
        # These will show up in sniff, accerciser, etc.
        self.headerbar.get_accessible().set_name("editor_headerbar")
        self.menu_button.get_accessible().set_name("main menu button")
        self.toplevel_widget.get_accessible().set_name("contents")
        self.mainhpaned.get_accessible().set_name("upper half")
        self.secondhpaned.get_accessible().set_name("tabs")
        self.main_tabs.get_accessible().set_name("primary tabs")
        self.context_tabs.get_accessible().set_name("secondary tabs")
        self.viewer.get_accessible().set_name("viewer")
        self.timeline_ui.get_accessible().set_name("timeline area")

        # Restore settings for position and visibility.
        if self.settings.mainWindowHPanePosition is None:
            self._set_default_positions()
        self.secondhpaned.set_position(self.settings.mainWindowHPanePosition)
        self.mainhpaned.set_position(self.settings.mainWindowMainHPanePosition)
        self.toplevel_widget.set_position(self.settings.mainWindowVPanePosition)

    def _set_default_positions(self):
        window_width = self.app.gui.get_size()[0]
        if self.settings.mainWindowHPanePosition is None:
            self.settings.mainWindowHPanePosition = window_width / 3
        if self.settings.mainWindowMainHPanePosition is None:
            self.settings.mainWindowMainHPanePosition = 2 * window_width / 3
        if self.settings.mainWindowVPanePosition is None:
            screen_width = float(self.app.gui.get_screen().get_width())
            screen_height = float(self.app.gui.get_screen().get_height())
            req = self.toplevel_widget.get_preferred_size()[0]
            if screen_width / screen_height < 0.75:
                # Tall screen, give some more vertical space the the tabs.
                value = req.height / 3
            else:
                value = req.height / 2
            self.settings.mainWindowVPanePosition = value

    def switch_context_tab(self, ges_clip):
        """Activates the appropriate tab on the second set of tabs.

        Args:
            ges_clip (GES.SourceClip): The clip which has been focused.
        """
        if isinstance(ges_clip, GES.TitleClip):
            page = 0
        elif isinstance(ges_clip, GES.SourceClip):
            page = 0
        elif isinstance(ges_clip, GES.TransitionClip):
            page = 1
        elif isinstance(ges_clip, GES.TestClip):
            page = 0
        else:
            self.warning("Unknown clip type: %s", ges_clip)
            return
        self.context_tabs.set_current_page(page)

    def focus_timeline(self):
        layers_representation = self.timeline_ui.timeline.layout
        # Check whether it has focus already, grab_focus always emits an event.
        if not layers_representation.props.is_focus:
            layers_representation.grab_focus()

    def __create_headerbar(self):
        headerbar = Gtk.HeaderBar()
        headerbar.set_show_close_button(True)

        undo_button = Gtk.Button.new_from_icon_name(
            "edit-undo-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        undo_button.set_always_show_image(True)
        undo_button.set_label(_("Undo"))
        undo_button.set_action_name("app.undo")
        undo_button.set_use_underline(True)

        redo_button = Gtk.Button.new_from_icon_name(
            "edit-redo-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        redo_button.set_always_show_image(True)
        redo_button.set_action_name("app.redo")
        redo_button.set_use_underline(True)

        # pylint: disable=attribute-defined-outside-init
        self.save_button = Gtk.Button.new_with_label(_("Save"))
        self.save_button.set_focus_on_click(False)

        self.render_button = Gtk.Button.new_from_icon_name(
            "system-run-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.render_button.set_always_show_image(True)
        self.render_button.set_label(_("Render"))
        self.render_button.set_tooltip_text(
            _("Export your project as a finished movie"))
        self.render_button.set_sensitive(False)  # The only one we have to set.
        self.render_button.connect("clicked", self._render_cb)

        undo_redo_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        undo_redo_box.get_style_context().add_class("linked")
        undo_redo_box.pack_start(undo_button, expand=False, fill=False, padding=0)
        undo_redo_box.pack_start(redo_button, expand=False, fill=False, padding=0)
        headerbar.pack_start(undo_redo_box)

        self.builder.add_from_file(
            os.path.join(get_ui_dir(), "mainmenubutton.ui"))

        self.menu_button = self.builder.get_object("menubutton")
        self.keyboard_shortcuts_button = self.builder.get_object("menu_shortcuts")

        headerbar.pack_end(self.menu_button)
        headerbar.pack_end(self.save_button)
        headerbar.pack_end(self.render_button)
        headerbar.show_all()

        return headerbar

    def _create_actions(self):
        group = Gio.SimpleActionGroup()
        self.toplevel_widget.insert_action_group("editor", group)
        self.headerbar.insert_action_group("editor", group)

        # pylint: disable=attribute-defined-outside-init
        self.save_action = Gio.SimpleAction.new("save", None)
        self.save_action.connect("activate", self.__save_project_cb)
        group.add_action(self.save_action)
        self.app.shortcuts.add("editor.save", ["<Primary>s"], self.save_action,
                               _("Save the current project"), group="win")
        self.save_button.set_action_name("editor.save")

        self.save_as_action = Gio.SimpleAction.new("save-as", None)
        self.save_as_action.connect("activate", self.__save_project_as_cb)
        group.add_action(self.save_as_action)
        self.app.shortcuts.add("editor.save-as", ["<Primary><Shift>s"],
                               self.save_as_action,
                               _("Save the current project as"), group="win")

        self.revert_to_saved_action = Gio.SimpleAction.new("revert-to-saved", None)
        self.revert_to_saved_action.connect("activate", self.__revert_to_saved_cb)
        group.add_action(self.revert_to_saved_action)

        self.export_project_action = Gio.SimpleAction.new("export-project", None)
        self.export_project_action.connect("activate", self.__export_project_cb)
        group.add_action(self.export_project_action)

        self.save_frame_action = Gio.SimpleAction.new("save-frame", None)
        self.save_frame_action.connect("activate", self.__save_frame_cb)
        group.add_action(self.save_frame_action)

        self.project_settings_action = Gio.SimpleAction.new("project-settings", None)
        self.project_settings_action.connect("activate", self.__project_settings_cb)
        group.add_action(self.project_settings_action)

        group.add_action(self.intro.intro_action)
        self.app.shortcuts.add("editor.interactive-intro", [], self.intro.intro_action,
                               _("Quick intros to Pitivi"), group="win")

        self.import_asset_action = Gio.SimpleAction.new("import-asset", None)
        self.import_asset_action.connect("activate", self.__import_asset_cb)
        group.add_action(self.import_asset_action)
        self.app.shortcuts.add("editor.import-asset", ["<Primary>i"],
                               self.import_asset_action,
                               _("Add media files to your project"), group="win")

    def __import_asset_cb(self, unused_action, unused_param):
        self.medialibrary.show_import_assets_dialog()

    def show_project_status(self):
        project = self.app.project_manager.current_project
        dirty = project.has_unsaved_modifications()
        self.save_action.set_enabled(dirty)
        self.revert_to_saved_action.set_enabled(bool(project.uri) and dirty)
        self.update_title()

# UI Callbacks

    def _media_library_play_cb(self, unused_medialibrary, asset):
        """Previews the specified asset.

        If the media library item to preview is an image, show it in the user's
        favorite image viewer. Else, preview the video/sound in Pitivi.
        """
        # Technically, our preview widget can show images, but it's never going
        # to do a better job (sizing, zooming, metadata, editing, etc.)
        # than the user's favorite image viewer.
        if asset.is_image():
            Gio.AppInfo.launch_default_for_uri(asset.get_id(), None)
        else:
            preview_window = PreviewAssetWindow(asset, self.app)
            preview_window.preview()

    def _project_changed_cb(self, unused_project):
        self.save_action.set_enabled(True)
        self.update_title()

# Toolbar/Menu actions callback

    def __save_project_cb(self, unused_action, unused_param):
        self.save_project()

    def __save_project_as_cb(self, unused_action, unused_param):
        self.save_project_as()

    def save_project(self):
        if not self.app.project_manager.current_project.uri or self.app.project_manager.disable_save:
            self.save_project_as()
        else:
            self.app.project_manager.save_project()

    def __revert_to_saved_cb(self, unused_action, unused_param):
        self.app.project_manager.revert_to_saved_project()

    def __export_project_cb(self, unused_action, unused_param):
        uri = self._show_export_dialog(self.app.project_manager.current_project)
        result = None
        if uri:
            result = self.app.project_manager.export_project(
                self.app.project_manager.current_project, uri)

        if not result:
            self.log("Project couldn't be exported")
        return result

    def __project_settings_cb(self, unused_action, unused_param):
        self.show_project_settings_dialog()

    def show_project_settings_dialog(self):
        project = self.app.project_manager.current_project
        dialog = ProjectSettingsDialog(self.app.gui, project, self.app)
        dialog.window.run()
        self.update_title()

# Project management callbacks

    def _project_manager_new_project_loaded_cb(self, project_manager, project):
        """Connects the UI to the specified project.

        Args:
            project_manager (ProjectManager): The project manager.
            project (Project): The project which has been loaded.
        """
        self.log("A new project has been loaded")

        self._connect_to_project(project)
        project.pipeline.activate_position_listener()

        self.viewer.set_project(project)
        self.clipconfig.set_project(project, self.timeline_ui)
        self.timeline_ui.set_project(project)

        # When creating a blank project there's no project URI yet.
        if project.uri:
            folder_path = os.path.dirname(path_from_uri(project.uri))
            self.settings.lastProjectFolder = folder_path

        self.update_title()

        if project_manager.disable_save is True:
            # Special case: we enforce "Save as", but the normal "Save" button
            # redirects to it if needed, so we still want it to be enabled:
            self.save_action.set_enabled(True)

        if project.ges_timeline.props.duration != 0:
            self.render_button.set_sensitive(True)

    def _project_manager_save_project_failed_cb(self, unused_project_manager, uri, exception=None):
        project_filename = unquote(uri.split("/")[-1])
        dialog = Gtk.MessageDialog(transient_for=self.app.gui,
                                   modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=_('Unable to save project "%s"') % project_filename)
        if exception:
            dialog.set_property("secondary-use-markup", True)
            dialog.set_property("secondary-text", unquote(str(exception)))
        dialog.set_transient_for(self.app.gui)
        dialog.run()
        dialog.destroy()
        self.error("failed to save project")

    def _project_manager_project_saved_cb(self, unused_project_manager, unused_project, unused_uri):
        self.update_title()
        self.save_action.set_enabled(False)

    def _project_manager_closing_project_cb(self, project_manager, project):
        """Investigates whether it's possible to close the specified project.

        Args:
            project_manager (ProjectManager): The project manager.
            project (Project): The project which has been closed.

        Returns:
            bool: True when it's OK to close it, False when the user chooses
                to cancel the closing operation.
        """
        if not project.has_unsaved_modifications():
            return True

        if project.uri and not project_manager.disable_save:
            save = _("Save")
        else:
            save = _("Save as...")

        dialog = Gtk.MessageDialog(transient_for=self.app.gui, modal=True)
        reject_btn = dialog.add_button(_("Close without saving"),
                                       Gtk.ResponseType.REJECT)

        dialog.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                           save, Gtk.ResponseType.YES)

        dialog.set_default_response(Gtk.ResponseType.CANCEL)
        dialog.get_accessible().set_name("unsaved changes dialog")
        reject_btn.get_style_context().add_class("destructive-action")

        primary = _("Save changes to the current project before closing?")
        dialog.props.use_markup = True
        dialog.props.text = "<span weight=\"bold\">" + primary + "</span>"

        if project.uri:
            path = unquote(project.uri).split("file://")[1]
            last_saved = max(
                os.path.getmtime(path), project_manager.time_loaded)
            time_delta = time() - last_saved
            message = _("If you don't save, "
                        "the changes from the last %s will be lost.") % \
                beautify_time_delta(time_delta)
        else:
            message = _("If you don't save, your changes will be lost.")

        dialog.props.secondary_text = message

        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.YES:
            if project.uri is not None and project_manager.disable_save is False:
                res = self.app.project_manager.save_project()
            else:
                res = self.save_project_as()
        elif response == Gtk.ResponseType.REJECT:
            res = True
        else:
            res = False

        return res

    def _project_manager_project_closed_cb(self, project_manager, project):
        """Starts disconnecting the UI from the specified project.

        This happens when the user closes the app or asks to load another
        project, immediately after the user confirmed that unsaved changes,
        if any, can be discarded but before the filechooser to pick the next
        project to load appears.
        """
        # We must disconnect from the project pipeline before it is released:
        if project.pipeline is not None:
            project.pipeline.deactivate_position_listener()

        self.info("Project closed")
        if project.loaded:
            self._disconnect_from_project(project)

        self.timeline_ui.set_project(None)
        self.clipconfig.set_project(None, None)

        self.render_button.set_sensitive(False)
        return False

    def _project_manager_reverting_to_saved_cb(self, unused_project_manager, unused_project):
        if self.app.project_manager.current_project.has_unsaved_modifications():
            dialog = Gtk.MessageDialog(transient_for=self.app.gui,
                                       modal=True,
                                       message_type=Gtk.MessageType.WARNING,
                                       buttons=Gtk.ButtonsType.NONE,
                                       text=_("Revert to saved project version?"))
            dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.NO,
                               Gtk.STOCK_REVERT_TO_SAVED, Gtk.ResponseType.YES)
            dialog.set_resizable(False)
            dialog.set_property("secondary-text",
                                _("This will reload the current project. All unsaved changes will be lost."))
            dialog.set_default_response(Gtk.ResponseType.NO)
            dialog.set_transient_for(self.app.gui)
            response = dialog.run()
            dialog.destroy()
            if response != Gtk.ResponseType.YES:
                return False
        return True

    def _project_manager_missing_uri_cb(self, project_manager, project, unused_error, asset):
        if project.at_least_one_asset_missing:
            # One asset is already missing so no point in spamming the user
            # with more file-missing dialogs, as we need all of them.
            return None

        if self.app.proxy_manager.is_proxy_asset(asset):
            uri = self.app.proxy_manager.get_target_uri(asset)
        else:
            uri = asset.get_id()

        dialog = MissingAssetDialog(self.app, asset, uri)
        new_uri = dialog.get_new_uri()

        if not new_uri:
            dialog.hide()
            if not self.app.proxy_manager.check_proxy_loading_succeeded(asset):
                # Reset the project manager and disconnect all the signals.
                project_manager.close_running_project()
                # Signal the project loading failure.
                # You have to do this *after* successfully creating a blank project,
                # or the startupwizard will still be connected to that signal too.
                reason = _("No replacement file was provided for \"<i>%s</i>\".\n\n"
                           "Pitivi does not currently support partial projects.") % \
                    info_name(asset)
                project_manager.emit("new-project-failed", project.uri, reason)

        dialog.destroy()
        return new_uri

    def _connect_to_project(self, project):
        project.connect("project-changed", self._project_changed_cb)
        project.ges_timeline.connect("notify::duration",
                                     self._timeline_duration_changed_cb)

    def _disconnect_from_project(self, project):
        project.disconnect_by_func(self._project_changed_cb)
        project.ges_timeline.disconnect_by_func(self._timeline_duration_changed_cb)

    def _timeline_duration_changed_cb(self, timeline, unused_duration):
        """Updates the render button.

        This covers the case when a clip is inserted into a blank timeline.
        This callback is not triggered by loading a project.
        """
        duration = timeline.get_duration()
        self.debug("Timeline duration changed to %s", duration)
        self.render_button.set_sensitive(duration > 0)

    def _show_export_dialog(self, project):
        self.log("Export requested")
        chooser = Gtk.FileChooserDialog(title=_("Export To..."),
                                        transient_for=self.app.gui,
                                        action=Gtk.FileChooserAction.SAVE)
        chooser.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                            _("Save"), Gtk.ResponseType.OK)
        chooser.set_default_response(Gtk.ResponseType.OK)

        chooser.set_select_multiple(False)
        chooser.props.do_overwrite_confirmation = True

        asset = GES.Formatter.get_default()
        asset_extension = asset.get_meta(GES.META_FORMATTER_EXTENSION)

        chooser.set_current_name(
            project.name + "." + asset_extension + "_tar")

        filt = Gtk.FileFilter()
        filt.set_name(_("Tar archive"))
        filt.add_pattern("*.%s_tar" % asset_extension)
        chooser.add_filter(filt)
        default = Gtk.FileFilter()
        default.set_name(_("Detect automatically"))
        default.add_pattern("*")
        chooser.add_filter(default)

        response = chooser.run()
        if response == Gtk.ResponseType.OK:
            self.log("User chose a URI to export project to")
            # need to do this to work around bug in Gst.uri_construct
            # which escapes all /'s in path!
            uri = "file://" + chooser.get_filename()
            self.log("uri: %s", uri)
            ret = uri
        else:
            self.log("User didn't choose a URI to export project to")
            ret = None

        chooser.destroy()
        return ret

    def save_project_as(self):
        uri = self._show_save_as_dialog()
        if uri is None:
            return False
        return self.app.project_manager.save_project(uri)

    def _show_save_as_dialog(self):
        self.log("Save URI requested")
        chooser = Gtk.FileChooserDialog(title=_("Save As..."),
                                        transient_for=self.app.gui,
                                        action=Gtk.FileChooserAction.SAVE)
        chooser.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                            _("Save"), Gtk.ResponseType.OK)
        chooser.set_default_response(Gtk.ResponseType.OK)
        asset = GES.Formatter.get_default()
        filt = Gtk.FileFilter()
        filt.set_name(asset.get_meta(GES.META_DESCRIPTION))
        filt.add_pattern("*.%s" % asset.get_meta(GES.META_FORMATTER_EXTENSION))
        chooser.add_filter(filt)

        chooser.set_select_multiple(False)
        chooser.set_current_name(_("Untitled") + "." +
                                 asset.get_meta(GES.META_FORMATTER_EXTENSION))
        chooser.set_current_folder(self.settings.lastProjectFolder)
        chooser.props.do_overwrite_confirmation = True

        default = Gtk.FileFilter()
        default.set_name(_("Detect automatically"))
        default.add_pattern("*")
        chooser.add_filter(default)

        response = chooser.run()
        if response == Gtk.ResponseType.OK:
            self.log("User chose a URI to save project to")
            # need to do this to work around bug in Gst.uri_construct
            # which escapes all /'s in path!
            uri = "file://" + chooser.get_filename()
            file_filter = chooser.get_filter().get_name()
            self.log("uri:%s , filter:%s", uri, file_filter)
            self.settings.lastProjectFolder = chooser.get_current_folder()
            ret = uri
        else:
            self.log("User didn't choose a URI to save project to")
            ret = None

        chooser.destroy()
        return ret

    def __save_frame_cb(self, unused_action, unused_param):
        """Exports a snapshot of the current frame as an image file."""
        res = self._show_save_screenshot_dialog()
        if res:
            path, mime = res[0], res[1]
            self.app.project_manager.current_project.pipeline.save_thumbnail(
                -1, -1, mime, path)

    def _show_save_screenshot_dialog(self):
        """Asks the user where to save the current frame.

        Returns:
            List[str]: The full path and the mimetype if successful, None otherwise.
        """
        chooser = Gtk.FileChooserDialog(title=_("Save As..."),
                                        transient_for=self.app.gui, action=Gtk.FileChooserAction.SAVE)
        chooser.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                            _("Save"), Gtk.ResponseType.OK)
        chooser.set_default_response(Gtk.ResponseType.OK)
        chooser.set_select_multiple(False)
        chooser.set_current_name(_("Untitled"))
        chooser.props.do_overwrite_confirmation = True
        formats = {_("PNG image"): ["image/png", ("png",)],
                   _("JPEG image"): ["image/jpeg", ("jpg", "jpeg")]}
        for image_format in formats:
            filt = Gtk.FileFilter()
            filt.set_name(image_format)
            filt.add_mime_type(formats.get(image_format)[0])
            chooser.add_filter(filt)
        response = chooser.run()
        if response == Gtk.ResponseType.OK:
            chosen_format = formats.get(chooser.get_filter().get_name())
            chosen_ext = chosen_format[1][0]
            chosen_mime = chosen_format[0]
            uri = os.path.join(
                chooser.get_current_folder(), chooser.get_filename())
            ret = ["%s.%s" % (uri, chosen_ext), chosen_mime]
        else:
            ret = None
        chooser.destroy()
        return ret

    def update_title(self):
        project = self.app.project_manager.current_project
        unsaved_mark = ""
        if project.has_unsaved_modifications():
            unsaved_mark = "*"
        title = "%s%s — %s" % (unsaved_mark, project.name, APPNAME)
        self.headerbar.set_title(title)


class PreviewAssetWindow(Gtk.Window):
    """Window for previewing a video or audio asset.

    Args:
        asset (GES.UriClipAsset): The asset to be previewed.
        app (Pitivi): The app.
    """

    def __init__(self, asset, app):
        Gtk.Window.__init__(self)
        self._asset = asset
        self.app = app

        self.set_title(_("Preview"))
        self.set_type_hint(Gdk.WindowTypeHint.UTILITY)
        self.set_transient_for(app.gui)

        self._previewer = PreviewWidget(app.settings, minimal=True)
        self.add(self._previewer)
        self._previewer.preview_uri(self._asset.get_id())
        self._previewer.show()

        self.connect("key-press-event", self._key_press_event_cb)

    def preview(self):
        """Shows the window and starts the playback."""
        width, height = self._calculate_preview_window_size()
        self.resize(width, height)
        # Setting the position of the window only works if it's currently hidden
        # otherwise, after the resize the position will not be readjusted
        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.show()

        self._previewer.play()
        # Hack so that we really really force the "utility" window to be
        # focused
        self.present()

    def _calculate_preview_window_size(self):
        info = self._asset.get_info()
        video_streams = info.get_video_streams()
        if not video_streams:
            # There is no video/image stream. This is an audio file.
            # Resize to the minimum and let the window manager deal with it.
            return 1, 1
        # For videos and images, automatically resize the window
        # Try to keep it 1:1 if it can fit within 85% of the parent window
        video = video_streams[0]
        img_width = video.get_natural_width()
        img_height = video.get_natural_height()
        mainwindow_width, mainwindow_height = self.app.gui.get_size()
        max_width = 0.85 * mainwindow_width
        max_height = 0.85 * mainwindow_height

        controls_height = self._previewer.bbox.get_preferred_size()[0].height
        if img_width < max_width and (img_height + controls_height) < max_height:
            # The video is small enough, keep it 1:1
            return img_width, img_height + controls_height
        else:
            # The video is too big, size it down
            # TODO: be smarter, figure out which (width, height) is bigger
            new_height = max_width * img_height / img_width
            return int(max_width), int(new_height + controls_height)

    def _key_press_event_cb(self, unused_widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.destroy()
        return True



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2009, Alessandro Decina <alessandro.d@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Undo/redo."""
import contextlib

from gi.repository import GObject

from pitivi.undo.base import ConditionsNotReadyYetError
from pitivi.undo.base import UndoableAction
from pitivi.undo.base import UndoError
from pitivi.undo.base import UndoWrongStateError
from pitivi.utils.loggable import Loggable


class UndoableActionStack(UndoableAction, Loggable):
    """A stack of UndoableAction objects.

    Attributes:
        action_group_name (str): The name of the operation.
        done_actions (List[UndoableAction]): The UndoableActions pushed in
            the stack.
        mergeable (bool): Whether this stack accepts merges with other
            compatible stacks.
        finalizing_action (FinalizingAction): The action to be performed
            at the end of undoing or redoing the stacked actions.
    """

    def __init__(self, action_group_name, mergeable, finalizing_action=None):
        UndoableAction.__init__(self)
        Loggable.__init__(self)
        self.action_group_name = action_group_name
        self.done_actions = []
        self.mergeable = mergeable
        self.finalizing_action = finalizing_action

    def __len__(self):
        return len(self.done_actions)

    def __repr__(self):
        return "%s: %s" % (self.action_group_name, self.done_actions)

    def attempt_merge(self, stack, action):
        """Merges the action into the previous one if possible.

        Returns:
            bool: Whether the merge has been done.
        """
        if not self.mergeable:
            return False

        if not self.done_actions:
            return False

        if not self.action_group_name == stack.action_group_name:
            return False

        return self.attempt_expand_action(action)

    def attempt_expand_action(self, action):
        """Expands the last action with the specified action if possible."""
        if not self.done_actions:
            return False

        last_action = self.done_actions[-1]
        return last_action.expand(action)

    def push(self, action):
        """Adds an action unless it's possible to expand the previous."""
        if self.attempt_expand_action(action):
            # The action has been merged into the last one.
            return

        self.done_actions.append(action)

    def _perform_actions(self, actions, method_name):
        delayed_actions = []
        delayed_reasons = {}
        for action in actions:
            self.log("Performing %s.%s()", action, method_name)
            method = getattr(action, method_name)
            try:
                method()
                for delayed_action in list(delayed_actions):
                    self.log("Performing delayed %s.%s()", action, method_name)
                    delayed_method = getattr(delayed_action, method_name)
                    try:
                        delayed_method()
                        self.log("Succeeded performing action")
                        delayed_actions.remove(delayed_action)
                        delayed_reasons.pop(delayed_action)
                    except ConditionsNotReadyYetError:
                        self.log("Further delaying action")
            except ConditionsNotReadyYetError as e:
                self.log("Delaying action")
                delayed_actions.append(action)
                delayed_reasons[action] = e
        if delayed_actions:
            raise UndoError("Delayable actions failed to apply: {} {}".format(delayed_actions, delayed_reasons))
        self.finish_operation()

    def do(self):
        self._perform_actions(self.done_actions, "do")

    def undo(self):
        self._perform_actions(self.done_actions[::-1], "undo")

    def finish_operation(self):
        if not self.finalizing_action:
            return
        self.finalizing_action.do()


class UndoableActionLog(GObject.Object, Loggable):
    """The undo/redo manager.

    A separate instance should be created for each Project instance.
    """

    __gsignals__ = {
        "begin": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "pre-push": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "push": (GObject.SignalFlags.RUN_LAST, None, (object, object)),
        "rollback": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "commit": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "move": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(self):
        GObject.Object.__init__(self)
        Loggable.__init__(self)

        self.undo_stacks = []
        self.redo_stacks = []
        self.stacks = []
        self.running = False
        self.rolling_back = False
        self._checkpoint = self._take_snapshot()

    @contextlib.contextmanager
    def started(self, action_group_name, **kwargs):
        """Gets a context manager which commits the transaction at the end."""
        self.begin(action_group_name, **kwargs)
        try:
            yield
        except:
            self.warning("An exception occurred while recording a "
                         "high-level operation. Rolling back.")
            self.rollback()
            raise
        else:
            self.commit(action_group_name)

    def begin(self, action_group_name, finalizing_action=None, mergeable=False, toplevel=False):
        """Starts recording a high-level operation which later can be undone.

        The recording can be stopped by calling the `commit` method or
        canceled by calling the `rollback` method.

        The operation will be composed of all the actions which have been
        pushed and also of the committed sub-operations.

        Args:
            action_group_name (str): The name of the operation.
            finalizing_action (FinalizingAction): The action to be performed
                at the end of undoing or redoing the stacked actions.
            mergeable (bool): Whether this stack accepts merges with future
                compatible stacks.
            toplevel (bool): If true, throws error if this operation is
                started while another one is being recorded.
        """
        if self.running:
            self.debug("Abort because running")
            return

        if toplevel and self.is_in_transaction():
            raise UndoWrongStateError("Toplevel operation started as suboperation", self.stacks)

        stack = UndoableActionStack(action_group_name, mergeable, finalizing_action)
        self.stacks.append(stack)
        self.debug("begin action group %s, nested %s",
                   stack.action_group_name, len(self.stacks))
        self.emit("begin", stack)

    def push(self, action):
        """Records a change noticed by the monitoring system.

        Args:
            action (Action): The action representing the change.
                If it's an UndoableAction, it's added to the current
                operation, if any.
        """
        self.emit("pre-push", action)

        if not isinstance(action, UndoableAction):
            # Nothing else to do with it.
            return

        if self.running:
            self.debug("Ignore push because running: %s", action)
            return

        if self.rolling_back:
            self.debug("Ignore push because rolling back: %s", action)
            return

        try:
            stack = self._get_last_stack()
        except UndoWrongStateError as e:
            self.warning("Failed pushing '%s' because no transaction started: %s", action, e)
            return

        merged = False
        if stack.mergeable and len(self.stacks[0]) == 0 and self.undo_stacks:
            # The current undoable operation is empty, this is the first action.
            # Check if it can be merged with the previous operation.
            previous_operation = self.undo_stacks[-1]
            if previous_operation.attempt_merge(stack, action):
                self.debug("Merging undoable operations")
                self.stacks = [self.undo_stacks.pop()]
                merged = True

        if not merged:
            stack.push(action)
            self.debug("push action %s in action group %s",
                       action, stack.action_group_name)

        self.emit("push", stack, action)

    def rollback(self, undo=True):
        """Forgets about the last started operation.

        Args:
            undo (bool): Whether to undo the last started operation.
                If False, it's disregarded without any action.
        """
        if self.running:
            self.debug("Ignore rollback because running")
            return

        self.debug("Rolling back, undo=%s", undo)
        self.rolling_back = True
        try:
            stack = self._get_last_stack(pop=True)
            self.debug("rollback action group %s, nested %s",
                       stack.action_group_name, len(self.stacks))
            self.emit("rollback", stack)
            if undo:
                stack.undo()
        finally:
            self.rolling_back = False

    def try_rollback(self, action_group_name):
        """Do rollback if the last started operation is @action_group_name."""
        try:
            last_operation = self._get_last_stack().action_group_name
        except UndoWrongStateError:
            return

        if last_operation == action_group_name:
            self.rollback()

    def commit(self, action_group_name):
        """Commits the last started operation."""
        if self.running:
            self.debug("Ignore commit because running")
            return

        self.debug("Committing %s", action_group_name)
        stack = self._get_last_stack(pop=True)
        if action_group_name != stack.action_group_name:
            raise UndoWrongStateError("Unexpected commit", action_group_name, stack, self.stacks)

        if not stack.done_actions:
            self.debug("Ignore empty stack %s", stack.action_group_name)
            return

        if not self.stacks:
            self.undo_stacks.append(stack)
            stack.finish_operation()
        else:
            self.stacks[-1].push(stack)

        if self.redo_stacks:
            self.redo_stacks = []

        self.debug("commit action group %s nested %s",
                   stack.action_group_name, len(self.stacks))
        self.emit("commit", stack)

    def undo(self):
        """Undoes the last recorded operation."""
        if self.stacks:
            raise UndoWrongStateError("Recording a transaction", self.stacks)
        if not self.undo_stacks:
            raise UndoWrongStateError("Nothing to undo")

        stack = self.undo_stacks.pop(-1)
        self.debug("Undo %s", stack)
        self._run(stack.undo)
        self.redo_stacks.append(stack)
        self.emit("move", stack)

    def redo(self):
        """Redoes the last undone operation."""
        if self.stacks:
            raise UndoWrongStateError("Recording a transaction", self.stacks)
        if not self.redo_stacks:
            raise UndoWrongStateError("Nothing to redo")

        stack = self.redo_stacks.pop(-1)
        self.debug("Redo %s", stack)
        self._run(stack.do)
        self.undo_stacks.append(stack)
        self.emit("move", stack)

    def _take_snapshot(self):
        return list(self.undo_stacks)

    def checkpoint(self):
        if self.stacks:
            raise UndoWrongStateError("Recording a transaction", self.stacks)

        self._checkpoint = self._take_snapshot()

    def dirty(self):
        current_snapshot = self._take_snapshot()
        return current_snapshot != self._checkpoint

    def _run(self, operation):
        self.running = True
        try:
            operation()
        finally:
            self.running = False

    def _get_last_stack(self, pop=False):
        try:
            if pop:
                stack = self.stacks.pop(-1)
            else:
                stack = self.stacks[-1]
        except IndexError as e:
            raise UndoWrongStateError("No transaction") from e

        return stack

    def is_in_transaction(self):
        """Gets whether currently recording an operation."""
        return bool(self.stacks)

    def has_assets_operations(self):
        """Checks whether user added/removed assets while working on the project."""
        for stack in self.undo_stacks:
            if stack.action_group_name in ["assets-addition", "assets-removal"]:
                return True
        return False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Utilities for getting the location of various directories.

Enables identical use for installed and uninstalled versions.
"""

import os.path
from gi.repository import GLib


# Again, mostly the same thing as in bin/pitivi.in and pitivi/utils/misc.py:
def _get_root_dir():
    return '/'.join(os.path.dirname(os.path.abspath(__file__)).split('/')[:-1])


def in_devel():
    """Returns whether the app is run from a git checkout."""
    return os.environ.get("PITIVI_DEVELOPMENT", "0") != "0"


LIBDIR = '/usr/lib'
PKGDATADIR = '/usr/share/pitivi'
GITVERSION = ''
VERSION = '2023.03'
APPNAME = 'Pitivi'
APPURL = 'https://www.pitivi.org/'
RELEASES_URL = 'https://www.pitivi.org/releases.txt'
APPMANUALURL_ONLINE = 'https://www.pitivi.org/manual/'
BUILDDIR = '/build/pitivi/src/build'


if in_devel():
    APPMANUALURL_OFFLINE = 'ghelp://%s' % os.path.join(_get_root_dir(), 'help/C')
else:
    APPMANUALURL_OFFLINE = 'help:pitivi'


def get_data_dir():
    if in_devel():
        datadir = os.path.join(_get_root_dir(), "data")
    elif os.getenv("PITIVI_TOP_LEVEL_DIR"):
        top_level = os.getenv("PITIVI_TOP_LEVEL_DIR")
        datadir = os.path.join(top_level, "data")
    else:
        datadir = PKGDATADIR
    return os.path.abspath(datadir)


def get_pixmap_dir():
    """Returns our directory with pixmaps."""
    return os.path.join(get_data_dir(), 'pixmaps')


def get_ui_dir():
    """Returns our directory with Gtk.Builder/Glade files."""
    return os.path.join(get_data_dir(), 'ui')


def get_audiopresets_dir():
    """Returns our directory with Audio Presets files."""
    return os.path.join(get_data_dir(), 'audiopresets')


def get_videopresets_dir():
    """Returns our directory with Video Presets files."""
    return os.path.join(get_data_dir(), 'videopresets')


def get_gstpresets_dir():
    """Returns our directory with Gst Presets files."""
    return os.path.join(get_data_dir(), 'gstpresets')


def get_plugins_dir():
    """Returns our default directory to store official plugins."""
    return os.path.join(_get_root_dir(), 'plugins')


def get_user_plugins_dir():
    """Returns our default directory to store non-official plugins."""
    user_data_dir = GLib.get_user_data_dir()
    user_plugin_dir = os.path.join(user_data_dir, 'pitivi', 'plugins')
    os.makedirs(user_plugin_dir, exist_ok=True)
    return user_plugin_dir



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2021, Piotr Brzeziński <thewildtreee@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import tempfile
import threading
from enum import Enum

from gi.repository import GES
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstPbutils
from gi.repository import GstTranscoder

from pitivi.utils.markers import MarkerListManager


class DetectionState(Enum):
    """Represents individual states of the beat detection process.

    Each state's value represents the percentage of completion in relation
    to the entire procedure.
    """

    IDLE = 0
    TRANSCODING = 25
    PREPARING_DETECTION = 50
    DETECTING = 75
    FINISHED = 100


class BeatDetector(GObject.Object):
    """Class responsible for performing beat detection on audio clips.

    Takes in a GESAudioSource, exposes methods for starting beat detection and
    clearing previously detected beats, as well as properties such as detection progress.
    Emits "detection-percentage" signals during beat detection and "detection-failed"
    when an error occurs during said process.

    Relies on GstTranscoder usage to extract the audio asset into raw WAV data
    and passes it to a beat detection library which returns timestamps of each detected beat.
    """

    __gsignals__ = {
        "detection-percentage": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "detection-failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(self, ges_source: GES.AudioSource):
        GObject.Object.__init__(self)
        self._ges_source = ges_source
        self._manager: MarkerListManager = ges_source.markers_manager
        self._state = DetectionState.IDLE

    @property
    def in_progress(self) -> bool:
        """Returns whether beat detection is currently in progress."""
        return DetectionState.IDLE.value < self.progress < DetectionState.FINISHED.value

    @property
    def progress(self) -> int:
        """Returns current percentage of completion of the beat detection process."""
        return self._state.value

    @property
    def beat_list_exists(self) -> bool:
        """Returns whether a marker list containing beat markers exists."""
        return self._manager.list_exists("beat_markers")

    def clear_beats(self):
        """Removes the marker list containing previously detected beats."""
        if self.in_progress or not self.beat_list_exists:
            return

        self._manager.remove_list("beat_markers")

    def detect_beats(self):
        """Starts beat detection on the asset given in the constructor.

        Will emit the "detection-percentage" signal during every step of the process.
        In case an error occurs, "detection-failed" will be emitted along with a string
        representing the cause of the failure.

        Will not do anything if beat detection is ongoing.
        """
        if self.in_progress:
            return

        GLib.idle_add(self._set_state, DetectionState.TRANSCODING)
        asset = self._ges_source.get_parent().get_asset()
        self.__perform_transcoding(asset)

    def _set_state(self, value):
        if self._state == value:
            return

        self._state = value
        self.emit("detection-percentage", self._state.value)

        # In case of the FINISHED state, we emit it
        # and then go back to idle without emitting.
        if self._state == DetectionState.FINISHED:
            self._state = DetectionState.IDLE

    def __perform_transcoding(self, asset: GES.Asset):
        profile = self.__create_encoding_profile()
        asset_uri = asset.get_id()

        result_file = tempfile.NamedTemporaryFile()  # pylint: disable=consider-using-with
        result_uri = Gst.filename_to_uri(result_file.name)
        transcoder = GstTranscoder.Transcoder.new_full(
            asset_uri, result_uri, profile)

        signals_emitter = transcoder.get_signal_adapter(None)
        # Passing transcoder here so the reference
        # doesn't get lost after the async call below.
        signals_emitter.connect("done", self.__transcoder_done_cb, result_file, transcoder)
        signals_emitter.connect(
            "error", self.__transcoder_error_cb, result_file, transcoder)

        transcoder.run_async()

    def __transcoder_done_cb(self, emitter, result_file, transcoder):
        self.__disconnect_transcoder(transcoder)
        thread = threading.Thread(target=self.__perform_beat_detection, args=(result_file,))
        thread.start()

    def __transcoder_error_cb(self, emitter, error, details, result_file, transcoder):
        result_file.close()
        self.__disconnect_transcoder(transcoder)
        GLib.idle_add(self._set_state, DetectionState.IDLE)
        self.emit("detection-failed", str(error))

    def __disconnect_transcoder(self, transcoder):
        signals_emitter = transcoder.get_signal_adapter(None)
        signals_emitter.disconnect_by_func(self.__transcoder_done_cb)
        signals_emitter.disconnect_by_func(self.__transcoder_error_cb)

    def __perform_beat_detection(self, audio_file):
        import librosa

        GLib.idle_add(self._set_state, DetectionState.PREPARING_DETECTION)
        y, sr = librosa.load(audio_file.name)

        GLib.idle_add(self._set_state, DetectionState.DETECTING)
        _, beat_times = librosa.beat.beat_track(y=y, sr=sr, units="time")

        audio_file.close()
        # Schedule marker creation in the UI thread.
        GLib.idle_add(self.__save_beat_markers, beat_times)

    def __save_beat_markers(self, beat_times):
        # Times from librosa are returned in seconds.
        marker_timestamps = [time * Gst.SECOND for time in beat_times]

        if self._manager.list_exists("beat_markers"):
            self._manager.remove_list("beat_markers")

        # Save the list, overwriting the current one if any.
        self._manager.add_list("beat_markers", marker_timestamps)
        self._manager.current_list_key = "beat_markers"
        self._set_state(DetectionState.FINISHED)

    def __create_encoding_profile(self):
        container_profile = GstPbutils.EncodingContainerProfile.new("wav-audio-profile",
                                                                    ("WAV audio-only container"),
                                                                    Gst.Caps("audio/x-wav"),
                                                                    None)
        audio_profile = GstPbutils.EncodingAudioProfile.new(
            Gst.Caps("audio/x-raw"), None, None, 0)
        container_profile.add_profile(audio_profile)
        return container_profile



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2021, Piotr Brzeziński <thewildtreee@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _
from typing import List
from typing import Optional
from typing import Tuple

from gi.repository import GES
from gi.repository import GObject

from pitivi.settings import GlobalSettings
from pitivi.timeline.markers import MarkersBox

# FIXME: Remove this once we depend on GES 1.20
GES_MARKERS_SNAPPABLE = hasattr(GES.MarkerList.new().props, "flags")

DEFAULT_LIST_KEY = "user_markers"
NAMES_DICT = {
    # Translators: The list of markers created by the user.
    DEFAULT_LIST_KEY: _("User markers"),
    # Translators: The list of markers representing detected audio beats.
    "beat_markers": _("Beats"),
}


class MarkerListManager(GObject.Object):
    """An abstraction layer between UI components and individual GESSources's marker lists.

    Attaches to a single GESSource, initialising a default marker list and tracking the
    addition / removal of any other ones. Keeps track of the currently active list, which
    is shown to the user if a MarkersBox has been attached via set_markers_box().
    """

    __gsignals__ = {
        # Emitted when a list is added or removed.
        "lists-modified": (GObject.SignalFlags.RUN_LAST, None, ()),
        # Emitted when the current list changes, along with its metadata key.
        "current-list-changed": (GObject.SignalFlags.RUN_LAST, None, (str,))
    }

    def __init__(self, settings: GlobalSettings, ges_source: GES.Source):
        GObject.Object.__init__(self)
        self._ges_source: Optional[GES.Source] = ges_source
        self._settings: GlobalSettings = settings
        self._box: Optional[MarkersBox] = None
        self._current_key: Optional[str] = None
        self.__ensure_default_list_exists()
        self._load_previous_or_default()
        self._set_default_snappability()

    def set_markers_box(self, markers_box: Optional[MarkersBox]):
        """Sets the MarkersBox used to display contents of the current marker list."""
        if self._box == markers_box:
            return

        if markers_box and self._ges_source != markers_box.ges_elem:
            raise ValueError("Marker box has to be attached to the same GESSource.")

        self._box = markers_box
        if not self._box:
            return

        self._box.markers_container = self.current_list

    def get_all_keys(self) -> List[str]:
        """Returns a list of metadata keys under which a marker list can be found."""
        list_keys = []
        self._ges_source.foreach(MarkerListManager.__clip_meta_foreach_func, list_keys)
        return list_keys

    def get_all_keys_with_names(self) -> List[Tuple[str, str]]:
        """Returns a list of list keys along with their human-readable names.

        This exists for ease of use with the MarkerProperties component.
        """
        list_keys = self.get_all_keys()
        # If no name is found just display the key directly.
        return [(key, NAMES_DICT.get(key, key)) for key in list_keys]

    def get_all_lists(self) -> List[GES.MarkerList]:
        """Returns a list of all existing marker lists."""
        list_keys = self.get_all_keys()
        return [self._ges_source.get_marker_list(key) for key in list_keys]

    def list_exists(self, list_key: str) -> bool:
        """Returns whether a list under the given metadata key exists."""
        return list_key in self.get_all_keys()

    def add_list(self, key: str, marker_timestamps: Optional[List[int]] = None) -> GES.MarkerList:
        """Creates an empty marker list and saves it under the given key.

        The key cannot be empty, can't contain spaces, and another list cannot
        already exist under the given key.
        A list of timestamps can be provided to automatically add corresponding
        markers to the newly created list.

        Emits the "lists-modified" signal after the list is successfully added.
        """
        if not key:
            raise ValueError("You must provide a key for the list.")
        if self._ges_source.get_marker_list(key):
            raise ValueError("A list already exists under the given key.")
        if " " in key:
            raise ValueError("List key cannot contain a space character.")

        marker_list = GES.MarkerList.new()

        if GES_MARKERS_SNAPPABLE and self._settings.markersSnappableByDefault:
            marker_list.props.flags |= GES.MarkerFlags.SNAPPABLE

        if marker_timestamps:
            for timestamp in marker_timestamps:
                marker_list.add(timestamp)

        self._ges_source.set_marker_list(key, marker_list)
        self.emit("lists-modified")
        return marker_list

    def remove_list(self, key: str):
        """Removes the marker list found under the given key.

        If the list being removed is currently active, the default list
        will be set as active instead.

        Note: the default "user_markers" cannot be removed.
        """
        if not key:
            raise ValueError("You must provide a key for the list.")

        if key == DEFAULT_LIST_KEY:
            raise ValueError("Cannot remove the default marker list.")

        if key == self.current_list_key:
            self._load_default()

        self._ges_source.set_meta(key, None)
        self.emit("lists-modified")

    @property
    def current_list_key(self) -> Optional[str]:
        """Returns the metadata key under which the current list can be found."""
        return self._current_key

    @current_list_key.setter
    def current_list_key(self, list_key: str):
        """Sets the marker list found under the given key as the currently active one.

        If no list should be active, an empty string needs to be given as the key.
        """
        if list_key is None:
            raise ValueError("No metadata key has been provided.")

        if self.current_list_key == list_key:
            return

        # Don't retrieve the list if the key is an empty string.
        new_list = None
        if list_key:
            new_list = self._ges_source.get_marker_list(list_key)
            if not new_list:
                raise ValueError("Invalid metadata key has been provided.")

        # Turn snappability off for lists going inactive.
        # The was_snappable value is not being serialized, thus
        # after reloading it will be set to the default user-preferred value.
        if self.current_list:
            self.current_list.was_snappable = self.snappable
            self.snappable = False

        # Set the new list as active and restore its snappable state if it's not None.
        self._current_key = list_key

        if new_list and hasattr(new_list, "was_snappable"):
            self.snappable = new_list.was_snappable
        if self._box:
            self._box.markers_container = new_list

        # This lets us preserve the current list between sessions.
        self._ges_source.set_string("last_chosen_list", list_key)
        self.emit("current-list-changed", self.current_list_key)

    @property
    def current_list(self) -> Optional[GES.MarkerList]:
        """Returns the currently active marker list."""
        if self.current_list_key is None:
            return None

        return self._ges_source.get_marker_list(self.current_list_key)

    @property
    def snappable(self) -> bool:
        """Returns whether the current list is considered a snapping target."""
        if not GES_MARKERS_SNAPPABLE:
            return False

        if not self.current_list:
            return False

        return self.current_list.props.flags & GES.MarkerFlags.SNAPPABLE

    @snappable.setter
    def snappable(self, snappable: bool):
        """Sets the snappable flag of the current list to a given value."""
        if not GES_MARKERS_SNAPPABLE:
            return

        if not self.current_list:
            return

        if self.snappable == snappable:
            return

        if snappable:
            self.current_list.props.flags |= GES.MarkerFlags.SNAPPABLE
        else:
            self.current_list.props.flags &= ~GES.MarkerFlags.SNAPPABLE

    def _load_previous_or_default(self):
        last_list_key = self._ges_source.get_string("last_chosen_list")

        # The default list is guaranteed to exist at this point.
        list_key = DEFAULT_LIST_KEY if last_list_key is None else last_list_key
        self.current_list_key = list_key

    def _load_default(self):
        self.current_list_key = DEFAULT_LIST_KEY

    def _set_default_snappability(self):
        """Sets user-chosen snappability state for all lists except for the active one.

        This is assumed to be only called once, after the default / previously chosen
        list has been loaded up.
        """
        snappable_by_default = self._settings.markersSnappableByDefault
        all_lists = self.get_all_lists()

        for marker_list in all_lists:
            # Ignore the active list - its flags were loaded from the project file.
            if marker_list == self.current_list:
                continue

            marker_list.was_snappable = snappable_by_default

    def __ensure_default_list_exists(self):
        if self._ges_source.get_marker_list(DEFAULT_LIST_KEY):
            return

        self.add_list(DEFAULT_LIST_KEY)

    @staticmethod
    def __clip_meta_foreach_func(container, key, value, keys):
        if isinstance(value, GES.MarkerList):
            keys.append(key)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Michael Westburg <michael.westberg@huskers.unl.edu>
# Copyright (c) 2020, Matt Lowe <mattlowe13@huskers.unl.edu>
# Copyright (c) 2020, Aaron Byington <aabyington4@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import cairo
from gi.repository import Gtk

from pitivi.utils.ui import gtk_style_context_get_color
from pitivi.utils.ui import NORMAL_FONT
from pitivi.utils.ui import set_cairo_color
from pitivi.utils.ui import SPACING

# The width for the peak meter
PEAK_METER_WIDTH = 8
# The maximum height for the peak meter
PEAK_METER_MAX_HEIGHT = 200
# The minimum height for the peak meter
PEAK_METER_MIN_HEIGHT = 80
# The number of cells on the peak meter bar
CELL_COUNT = 20
# The minimum peak value represented by the peak meter
MIN_PEAK = -60
# The font size for the scale
FONT_SIZE = 13
# The number of values shown on the scale
SCALE_COUNT = 5


class PeakMeterWidget(Gtk.DrawingArea):
    """Base class for peak meter components."""

    def __init__(self):
        Gtk.DrawingArea.__init__(self)
        self.pixel_buffer = None

        style_context = self.get_style_context()
        style_context.add_class("background")
        self.connect("size-allocate", self.__size_allocate_cb)

    def __size_allocate_cb(self, unused_event, allocation):
        if self.pixel_buffer is not None:
            self.pixel_buffer.finish()
            self.pixel_buffer = None

        self.pixel_buffer = cairo.ImageSurface(cairo.FORMAT_ARGB32, allocation.width, allocation.height)

    def draw_background(self, context, width, height):
        style_context = self.get_style_context()
        Gtk.render_background(style_context, context, 0, 0, width, height)


class PeakMeter(PeakMeterWidget):
    """A meter that shows peak values."""

    def __init__(self):
        PeakMeterWidget.__init__(self)
        self.peak = MIN_PEAK
        self.background_gradient = None
        self.peak_gradient = None
        width = PEAK_METER_WIDTH
        self.set_property("width_request", width)

        style_context = self.get_style_context()
        style_context.add_class("frame")

        self.connect("size-allocate", self.__size_allocate_cb)

    def do_draw(self, context):
        if self.pixel_buffer is None:
            return

        width = self.get_allocated_width()
        height = self.get_allocated_height()
        pixel_buffer = self.pixel_buffer

        drawing_context = cairo.Context(pixel_buffer)
        self.draw_background(drawing_context, width, height)
        self.__draw_bar(drawing_context, width, height)
        self.__draw_cells(drawing_context, width, height)
        self.__draw_frame(drawing_context, width, height)
        pixel_buffer.flush()

        context.set_source_surface(pixel_buffer, 0.0, 0.0)
        context.paint()

    def __size_allocate_cb(self, unused_event, allocation):
        self.__set_gradients(allocation.width, allocation.height)

    def __draw_bar(self, context, width, height):
        peak_height = self.__normalize_peak(height)

        context.set_source(self.background_gradient)
        context.rectangle(0, 0, width, height)
        context.fill()

        context.set_source(self.peak_gradient)
        context.rectangle(0, height - peak_height + 0, width, peak_height)
        context.fill()

    def __draw_cells(self, context, width, height):
        context.set_source_rgba(0.0, 0.0, 0.0, 0.5)
        context.set_line_width(2.0)

        cell_size = height / CELL_COUNT
        for i in range(1, CELL_COUNT):
            context.move_to(0, cell_size * i)
            context.line_to(width, cell_size * i)

        context.stroke()

    def __draw_frame(self, context, width, height):
        style_context = self.get_style_context()
        Gtk.render_frame(style_context, context, 0, 0, width, height)

    def __set_gradients(self, width, height):
        self.background_gradient = cairo.LinearGradient(0, 0, width, height)
        self.background_gradient.add_color_stop_rgb(1.0, 0.0, 0.3, 0.0)
        self.background_gradient.add_color_stop_rgb(0.7, 0.3, 0.3, 0.0)
        self.background_gradient.add_color_stop_rgb(0.0, 0.3, 0.0, 0.0)

        self.peak_gradient = cairo.LinearGradient(0, 0, width, height)
        self.peak_gradient.add_color_stop_rgb(1.0, 0.0, 1.0, 0.0)
        self.peak_gradient.add_color_stop_rgb(0.7, 1.0, 1.0, 0.0)
        self.peak_gradient.add_color_stop_rgb(0.0, 1.0, 0.0, 0.0)

    def __normalize_peak(self, height):
        return height / (-MIN_PEAK) * (max(self.peak, MIN_PEAK) - MIN_PEAK)

    def update_peakmeter(self, peak):
        self.peak = peak
        self.queue_draw()


class PeakMeterScale(PeakMeterWidget):
    """A scale for the peak meter."""

    def __init__(self):
        PeakMeterWidget.__init__(self)
        width = FONT_SIZE * 2
        self.set_property("width_request", width)

    def do_draw(self, context):
        if self.pixel_buffer is None:
            return

        width = self.get_allocated_width()
        height = self.get_allocated_height()
        pixel_buffer = self.pixel_buffer

        drawing_context = cairo.Context(pixel_buffer)
        self.draw_background(drawing_context, width, height)
        self.__draw_scale(drawing_context)
        pixel_buffer.flush()

        context.set_source_surface(pixel_buffer, 0.0, 0.0)
        context.paint()

    def __draw_scale(self, context):
        bar_height = self.get_bar_height()
        section_height = bar_height / (SCALE_COUNT - 1)

        style_context = self.get_style_context()
        color = gtk_style_context_get_color(style_context, Gtk.StateFlags.NORMAL)

        set_cairo_color(context, color)
        context.set_font_size(FONT_SIZE)
        context.set_font_face(NORMAL_FONT)
        text_extent = context.text_extents("0")

        for i in range(SCALE_COUNT):
            context.move_to(0, section_height * i + SPACING + text_extent.height / 2)
            context.show_text(str((MIN_PEAK // (SCALE_COUNT - 1)) * i))

    def get_bar_height(self):
        return self.get_allocated_height() - SPACING * 2
